"""
spotilyzer requests CSV
"""

# system imports
import csv

# project imports
from ..json.requests import REQUESTS_KEY, POD_NAME_KEY, REPLICAS_KEY, \
    CORE_LIMIT_KEY, MEM_LIMIT_KEY

# constants
_types = (str, int, float, float)


def load_requests(frequests):
    """
    Load requests CSV file. Formatting checks are performed
    :param frequests: path to CSV file.
    :return: requests object
    """
    request_list = []
    with open(frequests) as fobj:
        reader = csv.reader(fobj)
        header = reader.__next__()
        _validate_header(header, frequests)
        rowlen = len(header)
        line = 2
        for row in reader:
            _validate_row(rowlen, row, frequests, line)
            request = _get_request(rowlen, header, row, frequests, line)
            request_list.append(request)
            line += 1
    _validate_pod_names(request_list, frequests)
    return {REQUESTS_KEY: request_list}


def _validate_header(header, frequests):
    if header != [POD_NAME_KEY, REPLICAS_KEY, CORE_LIMIT_KEY, MEM_LIMIT_KEY]:
        raise SyntaxError(f"invalid header in {frequests}")


def _validate_row(rowlen, row, frequests, line):
    if len(row) != rowlen:
        raise SyntaxError("incorrect number of entries in "
                          f"{frequests}, line {line}")


def _get_request(rowlen, header, row, frequests, line):
    try:
        return {header[i]: _types[i](row[i]) for i in range(rowlen)}
    except ValueError as err:
        raise SyntaxError(f"invalid type in {frequests}, line {line}") from err


def _validate_pod_names(request_list, frequests):
    if len(request_list) != len(set(r[POD_NAME_KEY] for r in request_list)):
        raise SyntaxError(f"pod names in {frequests} are not unique")



"""
spotilyzer cli
"""

# system imports
import argparse
import os

# project imports
from .subcommands import subcommands

# constants
SUBCOMMAND_ARG = 'subcommand'


def parse_cmdline(argv):
    """
    Convert command line arguments to a argparse Namespace. Load subparsers
    from each subcommand.
    :param argv: list of command line arguments
    :return: argparse Namespace
    """
    parser = argparse.ArgumentParser(description="AWS spot fleet analyzer")
    subparsers = parser.add_subparsers(
        dest=SUBCOMMAND_ARG,
        description=f"use {os.path.basename(argv[0])} <subcommand> -h for "
                    "subcommand details"
    )
    for subcommand in subcommands:
        subcommand.add_parser(subparsers)
    return parser.parse_args(argv[1:])



"""
spotilyzer stats utilities
"""

# system imports
import math


class StatCollector:
    """
    Class to collect data for statistical analysis.
    """

    def __init__(self):
        self.count = 0
        self.total = 0.0
        self.square_total = 0.0

    def add(self, value):
        """
        Add data point to population.
        :param value: data value
        :return: None
        """
        self.count += 1
        self.total += value
        self.square_total += value * value

    def weighted_add(self, value, weight):
        """
        Add data point with weight to population
        :param value: data value
        :param weight: weight to apply to data value
        :return: None
        """
        self.count += weight
        self.total += weight * value
        self.square_total += weight * value * value

    def eval(self, sampled=True):
        """
        Compute average and standard deviation. Values are cached so any data
        added with subsequent add() calls are not used.
        :param sampled: Flag to indicate if data is sampled. A True value
        computes standard deviation with an n-1 denominator. A False value
        computes standard deviation with an n denominator.
        :return: tuple with average, standard deviation
        """
        if not hasattr(self, 'stats'):
            if self.count == 0:
                raise ZeroDivisionError("no data")
            avg = self.total / self.count
            if sampled:
                if self.count == 1:
                    raise ZeroDivisionError("insufficient data")
                variance = (self.square_total - self.total * avg) \
                           / (self.count - 1)
            else:
                variance = self.square_total / self.count - avg * avg
            if variance < 0.0:
                variance = 0.0
            self.stats = (avg, math.sqrt(variance))
        return self.stats


def quality(stat_collector, cov_level):
    """
    Determine if variance is within an acceptable level.
    :param stat_collector: StatCollector object
    :param cov_level: acceptable limit for coefficient of variation
    :return: True if coefficient of variation is within acceptable limit
    """
    avg, stddev =  stat_collector.eval(sampled=False)
    return stddev / avg < cov_level



"""
spotilyzer cli utilities
"""

# system imports
import argparse


def options(name, shortname=None):
    """
    Return options for argparse.
    :param name: long name
    :param shortname: short name to override first character of long name
    :return: tuple with short and long options
    """
    return f'-{shortname is None and name[0] or shortname}', \
           f'--{name}'


def posint(string):
    """
    Positive definite integer for argparse type.
    :param string: string value
    :return: integer value
    """
    try:
        value = int(string)
    except ValueError as err:
        raise argparse.ArgumentTypeError(err)
    if value < 1:
        raise argparse.ArgumentTypeError(f"invalid value {string}: value must "
                                         "be greater than 0")
    return value


def posfloat(string):
    """
    Positive semi-definite float type for argparse type.
    :param string: string value
    :return: float value
    """
    try:
        value = float(string)
    except ValueError as err:
        raise argparse.ArgumentTypeError(err)
    if value < 0.0:
        raise argparse.ArgumentTypeError(f"invalid value {string}: value must "
                                         "be greater than or equal to 0.0")
    return value


def csv(string):
    """
    Split CSV to a list.
    :param string: comma-separated list
    :return: list of values
    """
    return string.split(',')



"""
spotilyzer size-fleets subcommand
"""

# system imports
import collections
import math

# 3rd party imports
import tabulate

# project imports
from .base import SubCommand
from .utils.cli import options, posfloat
from .json.candidates import load_candidates, GROUPS_KEY, GROUP_NAME_KEY, \
    CANDIDATES_KEY, CANDIDATE_NAME_KEY, PRICE_KEY, RESOURCES_KEY, \
    MIN_CORES_KEY, MIN_MEMORY_KEY, AVG_CORES_KEY, AVG_MEMORY_KEY, REGION_KEY, \
    META_DATA_KEY, TIMESTAMP_KEY, DURATION_KEY
from .json.requests import load_requests, REPLICAS_KEY, REQUESTS_KEY, \
    CORE_LIMIT_KEY, MEM_LIMIT_KEY

# constants
_DESCRIPTION = "size fleets based on a request"
_BUFFER_OPT = 'buffer'
_BUFFER_DEFAULT = 20
_CANDIDATES_ARG = 'candidates'
_REQUESTS_ARG = 'requests'
_OUTPUT_ARG = 'output'
_TOTAL_CORES_KEY = 'total-cores'
_TOTAL_MEM_KEY = 'total-mem'
_MAX_CORES_KEY = 'max-cores'
_MAX_MEM_KEY = 'max-mem'
_INSTANCE_COUNT = 'instance-count'
_TOTAL_COST = 'total-cost'
_TABLE_HEADER = ('Group', 'Fleet', 'Instance Count', 'Total Cost')
_TABLE_FORMAT = 'grid'
_TABLE_COL_ALIGN = ('left', 'left', 'right', 'right')


class SizeFleets(SubCommand):
    """
    size-fleets subcommand
    """

    name = 'size-fleets'

    @classmethod
    def add_parser(cls, subparsers):
        """
        Add size-fleets subcommand parser.
        :param subparsers: object to attach parser
        :return: None
        """
        parser = subparsers.add_parser(cls.name, description=_DESCRIPTION,
                                       help=_DESCRIPTION)
        parser.add_argument(*options(_BUFFER_OPT), type=posfloat,
                            default=_BUFFER_DEFAULT,
                            help="percentage buffer on request (default is "
                                 f"{_BUFFER_DEFAULT}%%)")
        parser.add_argument(_CANDIDATES_ARG, help="JSON file with candidates")
        parser.add_argument(_REQUESTS_ARG,
                            help="file containing grouped requests")

    def run(self):
        """
        Generate instance count and price estimates from candidates and
        requests.
        :return: None
        """
        groups, candidates = self._get_groups()
        requirements = self._get_requirements()
        results = self._size_fleets(groups, requirements)
        self._display_results(results, candidates)

    def _get_groups(self):
        groups = {}
        candidates = load_candidates(self.getarg(_CANDIDATES_ARG))
        for group in candidates[GROUPS_KEY]:
            fleets = groups[group[GROUP_NAME_KEY]] = {}
            for candidate in group[CANDIDATES_KEY]:
                fleets[candidate[CANDIDATE_NAME_KEY]] = {
                    PRICE_KEY: candidate[PRICE_KEY],
                    RESOURCES_KEY: candidate[RESOURCES_KEY]
                }
        return groups, candidates

    def _get_requirements(self):
        requirements = collections.defaultdict(
            lambda: {
                _TOTAL_CORES_KEY: 0.0,
                _TOTAL_MEM_KEY: 0.0,
                _MAX_CORES_KEY: 0.0,
                _MAX_MEM_KEY:  0.0
            }
        )
        requests = load_requests(self.getarg(_REQUESTS_ARG),
                                 require_group=True)[REQUESTS_KEY]
        for request in requests:
            replicas = request[REPLICAS_KEY]
            cores = request[CORE_LIMIT_KEY]
            mem = request[MEM_LIMIT_KEY]
            requirement = requirements[request[GROUP_NAME_KEY]]
            requirement[_TOTAL_CORES_KEY] += replicas * cores
            requirement[_TOTAL_MEM_KEY] += replicas * mem
            requirement[_MAX_CORES_KEY] = max(cores,
                                              requirement[_MAX_CORES_KEY])
            requirement[_MAX_MEM_KEY] = max(mem, requirement[_MAX_MEM_KEY])
        return requirements

    def _size_fleets(self, groups, requirements):
        results = {}
        buffer = 1 + self.getarg(_BUFFER_OPT) / 100
        for group_name, fleets in groups.items():
            self._set_fleets(results, buffer, requirements, group_name, fleets)
        return results

    @staticmethod
    def _set_fleets(results, buffer, requirements, group_name, fleets):
        group_data = results[group_name] = {}
        requirement = requirements[group_name]
        required_cores = buffer * requirement[_MAX_CORES_KEY]
        required_mem = buffer * requirement[_MAX_MEM_KEY]
        total_cores = buffer * requirement[_TOTAL_CORES_KEY]
        total_mem = buffer * requirement[_TOTAL_MEM_KEY]
        for fleet_name, fleet_data in fleets.items():
            resources = fleet_data[RESOURCES_KEY]
            if required_cores > resources[MIN_CORES_KEY] \
                    or required_mem > resources[MIN_MEMORY_KEY]:
                continue
            count = max(
                math.ceil(total_cores / resources[AVG_CORES_KEY]),
                math.ceil(total_mem / resources[AVG_MEMORY_KEY])
            )
            group_data[fleet_name] = {
                _INSTANCE_COUNT: count,
                _TOTAL_COST: count * fleet_data[PRICE_KEY]
            }

    def _display_results(self, results, candidates):
        sorted_fleets = {
            group_name: sorted(
                fleets.keys(), key=lambda k: fleets[k][_TOTAL_COST]
            )
            for group_name, fleets in results.items()
        }
        print()
        print(f"{REGION_KEY}: {candidates[REGION_KEY]}")
        print(f"{META_DATA_KEY}:")
        print(f"  {TIMESTAMP_KEY}: {candidates[META_DATA_KEY][TIMESTAMP_KEY]}")
        print(f"  {DURATION_KEY}: {candidates[META_DATA_KEY][DURATION_KEY]} "
              "days")
        print(f"  {_BUFFER_OPT}: {self.getarg(_BUFFER_OPT)}%")
        print()
        table = [
            [
                group_name,
                '\n'.join(fleet_names),
                '\n'.join(
                    f"{results[group_name][f][_INSTANCE_COUNT]}"
                    for f in fleet_names
                ),
                '\n'.join(
                    f"{results[group_name][f][_TOTAL_COST]:.3f}"
                    for f in fleet_names
                )
            ]
            for group_name, fleet_names in sorted_fleets.items()
        ]
        print(tabulate.tabulate(table, headers=_TABLE_HEADER,
                                tablefmt=_TABLE_FORMAT,
                                colalign=_TABLE_COL_ALIGN))



"""
spotilyzer entry point
"""

# system imports
import sys

# project imports
from .cli import parse_cmdline
from .driver import run


def main(argv=sys.argv):
    """
    Entry point for spotilyzer. All exceptions are caught displayed and
    returned as a failing exit code.
    :param argv: command line arguments
    :return: exit code
    """
    try:
        args = parse_cmdline(argv)
        run(args)
    except SystemExit as exit_code:
        return exit_code
    except Exception as err:
        print(f"[error]: {err}")
        return 1
    return 0



"""
spotilyzer create-candidates subcommand
"""

# system imports
import collections
import datetime
import heapq
import sys
import textwrap

# 3rd party imports
import tabulate

# project imports
from .base import SubCommand
from .utils.cli import options, posint, csv
from .utils.paths import find_seeds_file
from .utils.stats import quality, StatCollector
from .json.seeds import load_seeds, SEEDS_KEY, GROUP_KEY, CANDIDATES_KEY, \
    CANDIDATE_NAME_KEY, INSTANCE_TYPES_KEY
from .json.candidates import save_candidates, REGION_KEY, GROUPS_KEY, \
    GROUP_NAME_KEY, MEM_CORE_RATIO_KEY, PRICE_KEY, RESOURCES_KEY, \
    AVG_CORES_KEY, AVG_MEMORY_KEY, MIN_CORES_KEY, MIN_MEMORY_KEY, \
    CANDIDATE_POOLS_KEY, INSTANCE_TYPE_KEY, AVAILABILITY_ZONES_KEY, \
    META_DATA_KEY, TIMESTAMP_KEY, DURATION_KEY, MINPOOL_KEY

# constants
_DESCRIPTION = "create candidate spot fleets from seeds"
_REGION_OPT = 'region'
_AVAILABILITY_ZONES_OPT = 'availability-zones'
_DURATION_OPT = 'duration'
_DURATION_DEFAULT = 90
_MINPOOL_OPT = 'minpool'
_MINPOOL_DEFAULT = 20
_SEEDS_OPT = 'seeds'
_CANDIDATES_ARG = 'candidates'
_SPOT_FLEETS_KEY = 'spot-fleets'
_AVG_MEM_CORE_RATIO_KEY = 'avg-mem-core-ratio'
_FLEETS_KEY = 'fleets'
_PRICE_KEY = 'price'
_MIN_STAT_COUNT = 30
_RESOURCES_KEY = 'resources'
_AVG_CORES_KEY = 'avg-cores'
_AVG_MEM_KEY = 'avg-mem'
_MIN_CORES_KEY = 'min-cores'
_MIN_MEM_KEY = 'min-mem'
_COV_LEVEL = 0.20
_MB_PER_GB = 1024
_TABLE_HEADER = ('Candidate', 'Price', 'Instance Type', 'Availability Zones')
_TABLE_FORMAT = 'grid'
_TABLE_FLOAT_FORMAT = '.3f'


class CreateCandidates(SubCommand):
    """
    create-candidates subcommand
    """

    name = 'create-candidates'

    @classmethod
    def add_parser(cls, subparsers):
        """
        Add create-candidates subcommand parser.
        :param subparsers: object to attach parser
        :return: None
        """
        parser = subparsers.add_parser(cls.name, description=_DESCRIPTION,
                                       help=_DESCRIPTION)
        parser.add_argument(*options(_REGION_OPT), default=None,
                            help="AWS region to query (default is from "
                                 "configuration or AWS_DEFAULT_REGION)")
        parser.add_argument(*options(_AVAILABILITY_ZONES_OPT), type=csv,
                            default=None,
                            help="comma-separated list of availability zones "
                                 "(defaults are all availability zones in the "
                                 "target region)")
        parser.add_argument(*options(_DURATION_OPT), type=posint,
                            default=_DURATION_DEFAULT,
                            help="number of days in price history (default "
                                 f"{_DURATION_DEFAULT})")
        parser.add_argument(*options(_MINPOOL_OPT), type=posint,
                            default=_MINPOOL_DEFAULT,
                            help="minimum number of pools in spot fleet "
                                 f"(default {_MINPOOL_DEFAULT})")
        parser.add_argument(*options(_SEEDS_OPT), default=None,
                            help="JSON file with seed instance types (if not "
                                 "specified use user's copy or the package "
                                 "data)")
        parser.add_argument(_CANDIDATES_ARG,
                            help="JSON file for candidates output")

    def run(self):
        """
        Create candidates from a seed file and spot price history data.
        :return: None
        """
        self._set_region()
        seeds = self._get_seeds()
        data = self._start(seeds)
        timestamp = self._collect_data(data)
        self._filter_data(data)
        candidates = self._select_candidates(data)
        self._price_candidates(candidates)
        self._get_candidate_resources(candidates)
        results = self._save_candidates(candidates, timestamp)
        self._display_results(results)

    def _set_region(self):
        region = self.getarg(_REGION_OPT)
        if region is not None:
            self.reconnect(region)

    def _get_seeds(self):
        seeds_file = self.getarg(_SEEDS_OPT)
        if seeds_file is None:
            seeds_file = find_seeds_file()
        return load_seeds(seeds_file)

    def _start(self, seeds):
        availability_zones = self.getarg(_AVAILABILITY_ZONES_OPT)
        if availability_zones is None:
            availability_zones = self._get_azs()
        else:
            self._validate_azs(availability_zones)
        return self._start_seed_data(seeds, availability_zones)

    def _get_azs(self):
        response = self.client.describe_availability_zones()
        return [a['ZoneName'] for a in response['AvailabilityZones']]

    def _validate_azs(self, availability_zones):
        valid_azs = set(self._get_azs())
        invalid_azs = set(availability_zones) - valid_azs
        if len(invalid_azs) > 0:
            raise ValueError("invalid availability zone(s) for region "
                             f"{self.client.meta.region_name}: "
                             f"{','.join(invalid_azs)}")

    def _start_seed_data(self, seeds, availability_zones):
        data = {}
        del_list = []
        for seed in seeds[SEEDS_KEY]:
            group_data = data[seed[GROUP_KEY]] = {}
            instance_type_map = self._map_instance_type_data(
                group_data,
                seed[CANDIDATES_KEY]
            )
            paginator \
                = self.client.get_paginator('describe_instance_type_offerings')
            for response in paginator.paginate(
                LocationType='availability-zone',
                Filters=[
                    {
                        'Name': 'location',
                        'Values': availability_zones
                    },
                    {
                        'Name': 'instance-type',
                        'Values': list(instance_type_map.keys())
                    }
                ]
            ):
                for offering in response['InstanceTypeOfferings']:
                    instance_type_map[
                        offering['InstanceType']
                    ][
                        offering['Location']
                    ] = StatCollector()
            group_del_list = [
                (c, i)
                for c, g in group_data.items()
                for i, a in g.items()
                if len(a) == 0
            ]
            for del_item in group_del_list:
                del group_data[del_item[0]][del_item[1]]
            del_list.extend(group_del_list)
        if len(del_list) > 0:
            print()
            print(
                textwrap.fill(
                    "[warning]: the following instance types are not "
                    "available in this region: "
                    f"{', '.join(d[1] for d in del_list)}"
                ), '\n'
            )
        return data

    @staticmethod
    def _map_instance_type_data(group_data, candidates):
        instance_type_map = {}
        for candidate in candidates:
            instance_type_data \
                = group_data[candidate[CANDIDATE_NAME_KEY]] = {}
            for instance_type in candidate[INSTANCE_TYPES_KEY]:
                instance_type_map[instance_type] \
                    = instance_type_data[instance_type] = {}
        return instance_type_map

    def _collect_data(self, data):
        print("collecting spot price data...")
        end = datetime.datetime.now()
        start = end - datetime.timedelta(days=self.getarg(_DURATION_OPT))
        az_map = collections.defaultdict(dict)
        for group_data in data.values():
            for instance_type_data in group_data.values():
                for instance_type, az_data in instance_type_data.items():
                    for azone, data_rec in az_data.items():
                        az_map[azone][instance_type] = data_rec
        paginator = self.client.get_paginator('describe_spot_price_history')
        for azone, az_map_data in az_map.items():
            for response in paginator.paginate(
                StartTime=start,
                EndTime=end,
                AvailabilityZone=azone,
                InstanceTypes=list(az_map_data.keys()),
                ProductDescriptions=["Linux/UNIX (Amazon VPC)"]
            ):
                for history_rec in response['SpotPriceHistory']:
                    az_map_data[history_rec['InstanceType']].add(
                        float(history_rec['SpotPrice'])
                    )
            print(f"  {azone}: {sum(d.count for d in az_map_data.values())} "
                  "records")
        return end

    def _filter_data(self, data):
        del_list = []
        for group_data in data.values():
            for instance_type_data in group_data.values():
                del_set = set(i
                    for i, ad in instance_type_data.items()
                    if any(d.count < _MIN_STAT_COUNT for d in ad.values())
                )
                for instance_type in del_set:
                    del instance_type_data[instance_type]
                del_list.extend(del_set)
            self._cleanup_dict(group_data)
        self._cleanup_dict(data)
        if len(data) == 0:
            raise RuntimeError("duration too short to collect any "
                               "statistically significant samples")
        if len(del_list) > 0:
            print()
            print(
                textwrap.fill(
                        "[warning]: removed the following instance types "
                        "because of insufficient data: "
                        f"{', '.join(del_list)}"
                ), '\n'
            )

    @staticmethod
    def _cleanup_dict(data_dict):
        del_list = [k for k,v in data_dict.items() if len(v) == 0]
        for key in del_list:
            del data_dict[key]

    def _select_candidates(self, data):
        print("selecting candidates...")
        minpool = self.getarg(_MINPOOL_OPT)
        candidates = {}
        for group, group_data in data.items():
            spot_candidates = {}
            candidates[group] = {
                _SPOT_FLEETS_KEY: spot_candidates,
                _AVG_MEM_CORE_RATIO_KEY: None
            }
            for candidate, instance_type_data in group_data.items():
                fleet_candidates = {}
                spot_candidates[candidate] = {
                    _FLEETS_KEY: fleet_candidates,
                    _RESOURCES_KEY: {},
                    _PRICE_KEY: None
                }
                if sum(len(a) for a in instance_type_data.values()) <= minpool:
                    for instance_type, az_data in instance_type_data.items():
                        fleet_candidates[instance_type] \
                            = self._convert_azdata(az_data)
                    continue
                self._select_nsmallest(fleet_candidates, instance_type_data,
                                       minpool)
        return candidates

    def _select_nsmallest(self, fleet_candidates, instance_type_data, minpool):
        heap = []
        for instance_type, az_data in instance_type_data.items():
            dyn_price = sum(
                sum(i.eval()) for i in az_data.values()
            ) / len(az_data)
            heapq.heappush(heap, (dyn_price, instance_type))
        pool_count = 0
        while pool_count < minpool:
            item = heapq.heappop(heap)
            az_data = instance_type_data[item[1]]
            fleet_candidates[item[1]] = self._convert_azdata(az_data)
            pool_count += len(az_data)

    @staticmethod
    def _convert_azdata(az_data):
        return {az: s.eval()[0] for az, s in az_data.items()}

    @staticmethod
    def _price_candidates(candidates):
        for candidate_record in candidates.values():
            for spot_candidate in candidate_record[_SPOT_FLEETS_KEY].values():
                total = sum(
                    s
                    for p in spot_candidate[_FLEETS_KEY].values()
                    for s in p.values()
                )
                count = sum(
                    len(p) for p in spot_candidate[_FLEETS_KEY].values()
                )
                spot_candidate[_PRICE_KEY] = total / count

    def _get_candidate_resources(self, candidates):
        print("getting candidate resources...")
        paginator = self.client.get_paginator('describe_instance_types')
        for group, candidate_record in candidates.items():
            mcstats = StatCollector()
            for candidate, spot_candidate in candidate_record[
                _SPOT_FLEETS_KEY
            ].items():
                self._compute_resource_stats(paginator, group, candidate,
                                             spot_candidate, mcstats)
            if not quality(mcstats, _COV_LEVEL):
                print(f"[warning]: memory/core ratios for {group} have a "
                      "large variance")
            candidate_record[_AVG_MEM_CORE_RATIO_KEY] = mcstats.eval()[0]

    @staticmethod
    def _compute_resource_stats(paginator, group, candidate, spot_candidate,
                                mcstats):
        cstats = StatCollector()
        mstats = StatCollector()
        core_min = sys.float_info.max
        mem_min = sys.float_info.max
        fleets = spot_candidate[_FLEETS_KEY]
        for response in paginator.paginate(
                InstanceTypes=list(fleets.keys())
        ):
            for instance_type_data in response['InstanceTypes']:
                cores = instance_type_data['VCpuInfo']['DefaultVCpus']
                mem = instance_type_data['MemoryInfo']['SizeInMiB']
                weight = len(
                    fleets[instance_type_data['InstanceType']]
                )
                cstats.weighted_add(cores, weight)
                mstats.weighted_add(mem, weight)
                mcstats.weighted_add(mem / cores, weight)
                core_min = min(cores, core_min)
                mem_min = min(mem, mem_min)
        if not quality(cstats, _COV_LEVEL):
            print(f"[warning]: cores for {group}/{candidate} have a "
                  "large variance")
        if not quality(mstats, _COV_LEVEL):
            print(f"[warning]: memories for {group}/{candidate} have "
                  "a large variance")
        spot_candidate[_RESOURCES_KEY].update(
            {
                _AVG_CORES_KEY: cstats.eval(sampled=False)[0],
                _AVG_MEM_KEY: mstats.eval(sampled=False)[0],
                _MIN_CORES_KEY: core_min,
                _MIN_MEM_KEY: mem_min
            }
        )

    def _save_candidates(self, candidates, timestamp):
        group_list = []
        results = {
            REGION_KEY: self.client.meta.region_name,
            META_DATA_KEY: {
                TIMESTAMP_KEY: timestamp.isoformat(),
                DURATION_KEY: self.getarg(_DURATION_OPT),
                MINPOOL_KEY: self.getarg(_MINPOOL_OPT)
            },
            GROUPS_KEY: group_list
        }
        for spot_fleet, candidate_record in candidates.items():
            candidate_list = []
            group_list.append(
                {
                    GROUP_NAME_KEY: spot_fleet,
                    MEM_CORE_RATIO_KEY: candidate_record[
                        _AVG_MEM_CORE_RATIO_KEY
                    ] / _MB_PER_GB,
                    CANDIDATES_KEY: candidate_list
                }
            )
            for candidate, fleet_candidates in candidate_record[
                _SPOT_FLEETS_KEY
            ].items():
                resources = fleet_candidates[_RESOURCES_KEY]
                pool_list = []
                candidate_list.append(
                    {
                        CANDIDATE_NAME_KEY: candidate,
                        PRICE_KEY: fleet_candidates[_PRICE_KEY],
                        RESOURCES_KEY: {
                            AVG_CORES_KEY: resources[_AVG_CORES_KEY],
                            AVG_MEMORY_KEY: resources[_AVG_MEM_KEY]
                                            / _MB_PER_GB,
                            MIN_CORES_KEY: resources[_MIN_CORES_KEY],
                            MIN_MEMORY_KEY: resources[_MIN_MEM_KEY]
                                            / _MB_PER_GB
                        },
                        CANDIDATE_POOLS_KEY: pool_list
                    }
                )
                for instance_type, instance_type_stats in fleet_candidates[
                    _FLEETS_KEY
                ].items():
                    pool_list.append(
                        {
                            INSTANCE_TYPE_KEY: instance_type,
                            AVAILABILITY_ZONES_KEY: list(
                                instance_type_stats.keys()
                            )
                        }
                    )
        save_candidates(self.getarg(_CANDIDATES_ARG), results)
        return results

    @staticmethod
    def _display_results(results):
        print()
        print(f"{REGION_KEY}: {results[REGION_KEY]}")
        print(f"{META_DATA_KEY}:")
        print(f"  {TIMESTAMP_KEY}: {results[META_DATA_KEY][TIMESTAMP_KEY]}")
        print(f"  {DURATION_KEY}: {results[META_DATA_KEY][DURATION_KEY]} days")
        print(f"  {MINPOOL_KEY}: {results[META_DATA_KEY][MINPOOL_KEY]}")
        for group in results[GROUPS_KEY]:
            print()
            print(f"group: {group[GROUP_NAME_KEY]}")
            table = [
                [
                    c[CANDIDATE_NAME_KEY],
                    c[PRICE_KEY],
                    '\n'.join(
                        p[INSTANCE_TYPE_KEY] for p in c[CANDIDATE_POOLS_KEY]
                    ),
                    '\n'.join(
                        ','.join(p[AVAILABILITY_ZONES_KEY])
                        for p in c[CANDIDATE_POOLS_KEY]
                    )
                ]
                for c in group[CANDIDATES_KEY]
            ]
            print(tabulate.tabulate(table, headers=_TABLE_HEADER,
                                    tablefmt=_TABLE_FORMAT,
                                    floatfmt=_TABLE_FLOAT_FORMAT))



"""
spotilyzer requests JSON
"""

# system imports
import json

# 3rd party imports
import jsonschema

# project imports
from .candidates import GROUP_NAME_KEY

# schema
REQUESTS_KEY = 'requests'
POD_NAME_KEY = 'pod-name'
REPLICAS_KEY = 'replicas'
CORE_LIMIT_KEY = 'core-limit'
MEM_LIMIT_KEY = 'mem-limit'
_SCHEMA = {
    'type': 'object',
    'properties': {
        REQUESTS_KEY: {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    POD_NAME_KEY: {
                        'type': 'string'
                    },
                    GROUP_NAME_KEY: {
                        'type': 'string'
                    },
                    REPLICAS_KEY: {
                        'type': 'integer',
                        'minimum': 1
                    },
                    CORE_LIMIT_KEY: {
                        'type': 'number',
                        'exclusiveMinimum': 0.0
                    },
                    MEM_LIMIT_KEY: {
                        'type': 'number',
                        'exclusiveMinimum': 0.0
                    }
                },
                'required': [
                    POD_NAME_KEY,
                    REPLICAS_KEY,
                    CORE_LIMIT_KEY,
                    MEM_LIMIT_KEY
                ],
                'additionalProperties': False
            },
            'additionalItems': False,
            'uniqueItems': True,
            'minItems': 1
        }
    },
    'required': [
        REQUESTS_KEY
    ],
    'additionalProperties': False
}

# constants
_JSON_INDENT = 2


def load_requests(frequests, require_group=False):
    """
    Load requests JSON file. Schema and consistency checks are performed.
    :param frequests: path to JSON file
    :param require_group: flag to require group key in JSON document
    :return: requests object
    """
    with open(frequests) as fobj:
        requests = json.load(fobj)
    jsonschema.validate(requests, _SCHEMA)
    _validate(requests, frequests, require_group)
    return requests


def save_requests(frequests, requests):
    """
    Save a requests object to a JSON file. Schema and consistency checks are
    performed before saving.
    :param frequests: path to output JSON file
    :param requests: requests object
    :return: None
    """
    jsonschema.validate(requests, _SCHEMA)
    _validate(requests, frequests, True)
    with open(frequests, 'w') as fobj:
        json.dump(requests, fobj, indent=_JSON_INDENT)


def _validate(requests, frequests, require_group):
    _requests = requests[REQUESTS_KEY]
    if len(_requests) != len(set(r[POD_NAME_KEY] for r in _requests)):
        raise SyntaxError(f"duplicate pod name in {frequests}")
    if require_group and any(GROUP_NAME_KEY not in r for r in _requests):
        raise SyntaxError(f"{GROUP_NAME_KEY} missing in request record")
